#! R

######################################################################################
# Script to create training data for the bootstrap sample data sets. 
#   Is similar to the RempelBootstrapTest.R except it will generate training data on 
#         the bootstrap samples (note, ensure that these samples used for training are not the same as used for testing)
#
# Assumes a .m script in the working directory titled: RempelTrainingModelScript.m
#   Inputs: BootstrapData.RData files generated from BootstrapSampling.R
#
# Output: .mat files containing the k-NN trained models 
#
######################################################################################

## Initialization

# Dependencies

FurfineScript <- RempelTrainingModelScript <-  'RempelTrainingModelScript' # matlab Training Matching for kNN variant
source('InitRmatlabFunctions.R')
source('RempelModelTraining_SlaveFunctions.R')

# Global Parameters
portnum <- 9999 # local host port number to be used for R - matlab connection
Seed <- 195264	
set.seed(Seed)
Nchunks = 1;
ModelTraining = 0;


# Script Parameters
args <- commandArgs(trailingOnly = TRUE)
inputfilename <- args[1]
outputfilename <- args[2]
methodTreatDuplicates <- args[3]
Parallel <- as.numeric(args[4])
nMatSlaves <- as.numeric(args[5])
min_sentime <- as.numeric(args[6])
min_loansize <- as.numeric(args[7])
bpsl <- as.numeric(args[8])
bpsh <- as.numeric(args[9])
refinement <- args[10]

if(length(args) > 10){
# then it is the additional options for Rempel Criterion
        SyntheticTraining <- as.numeric(args[11])
        ControlSyntheticForParticipants <- as.numeric(args[12])
        useDANN <- as.numeric(args[13])
        klossWeighted <- as.numeric(args[14])
        choiceK <- as.numeric(args[15])
        Xpaythreshold <- as.numeric(args[16])
        methodDup2 <- args[17]
        training_suffix_str <- args[18]
	furfineData_filename <- args[19]
	useReturnLegCharacteristics <- 1
}

# Check that Matched Data exists
stopifnot(file.exists(furfineData_filename))
jobname <- sub('.RData', '', outputfilename)
Simulation <- 1
load(inputfilename)
numboots <- length(BootSample)
load(furfineData_filename) # furfdat object
head(furfdat)
suppressWarnings(subsample_number <- as.numeric(substr(inputfilename, regexpr('[AB]?.RData$', inputfilename) -2, regexpr('[AB]?.RData$', inputfilename)-1)))
if(is.na(subsample_number)) subsample_number <- as.numeric(substr(inputfilename, regexpr('[AB]?.RData$', inputfilename) -1, regexpr('[AB]?.RData$', inputfilename)-1)) # is a one digit number
print(paste('subsample number ', subsample_number))


###########################################################################


# Check if Rmpi set-up

num_cores <- NULL
suppressWarnings(try({require(Rmpi)}, silent=T))
try({mpi.spawn.Rslaves()}, silent=T)
try({num_cores <- mpi.comm.size()},silent=T )

if(is.null(num_cores) || num_cores <= 1){
######################################
	# Serial computation
        rank <- 0
        OUT <- try({openLocalRmatlabConnection(port=portnum)}, silent=T)
        try({success <- OUT$success}, silent=T)
        try({matlab <- OUT$matlab}, silent=T)


        if(success==1 || inherits(OUT,'try-error')){
                close(matlab)
                stop('Failed to properly open matlab connection. Aborting.')

        } else {
                print("Succesfully opened R-matlab connection.")
        }

       setVariable(matlab, Seed=Seed, rank=rank)
	evaluate(matlab, 'rng(Seed);')
       setVariable(matlab, Nchunks=Nchunks,
                Parallel=Parallel, nMatSlaves=nMatSlaves, min_sentime=min_sentime,
                min_loansize=min_loansize, methodTreatDuplicates=methodTreatDuplicates,
                refinement=refinement,
                output_string = outputfilename
        )
        if(length(args) > 10){
            setVariable(matlab, Xpaythreshold=Xpaythreshold,
                    choiceK=choiceK,
                    ControlSyntheticForParticipants=ControlSyntheticForParticipants,
                    useReturnLegCharacteristics=useReturnLegCharacteristics,
                    useDANN=useDANN,
                    klossWeighted=klossWeighted,
                    ModelTraining=ModelTraining,
                    Simulation=Simulation,
                    methodDup2=methodDup2,
                    subsample_number=subsample_number
            )
        }

        evaluate(matlab,
        "fprintf('[%d] number of matlab slave: %d', rank, nMatSlaves);")
	evaluate(matlab,
        "fprintf('[%d] parameters uploaded', rank);")
        ## Starting
	evaluate(matlab,
                "addpath('KNN');")
        # set seed for each matlab worker?
        evaluate(matlab,
                "fprintf('[%d] subsample number: %d', rank, subsample_number);")
        print("Main Function About To Start")
	tasks <- lapply(1:numboots, function(x) list(bootNumber=x, bootsample=BootSample[[x]]$bootsample))
	Results <- lapply(tasks, function(x) RempelModelFunction(x, furfdat))
	# Note nothing is outputted to R, results are saved in .mat files	
	print('Model Training Procedure Complete')
	close(matlab)
######################################	
} else {
######################################
	# Parallel computation

	# 
	init.slave <- function(){
                    out_fname <- sprintf("BootstrapSamplingAlgorithm_output_slave.%03d.txt", mpi.comm.rank(0))
                    zz <- file( out_fname, open="wt" )
                    sink( zz, append=TRUE )
                    sink( zz, type="message", append=TRUE )
        }
	mpi.bcast.Robj2slave(init.slave)
	mpi.bcast.cmd(init.slave())
	
	# Open R-matlab connection on slaves
        mpi.bcast.Robj2slave(openMatlabConnection)
        success <- mpi.remote.exec(try({openMatlabConnection(jobname, mpi.comm.rank())}), ret=TRUE)

        if (any(sapply(success, inherits, "try-error")) || !all(as.logical(success))) {
            mpi.bcast.cmd(print('Failed to properly open matlab connection. Aborting.'))
            mpi.bcast.cmd( close(matlab) )
            mpi.close.Rslaves()
            mpi.quit()
        } else {
                mpi.bcast.cmd( print("Succesfully opened matlab connection."))
        }
	
	# setup independent streams for each worker, sample seed to set for Matlab worker
        mpi.setup.rngstream(iseed=Seed)
        mpi.bcast.cmd(seed <- sample(3*10^7:7*10^7,1))

	# distribute parameters to workers
	mpi.bcast.cmd(setVariable(matlab, seed=seed))
	mpi.bcast.cmd(evaluate(matlab, 'rng(seed);'))
	mpi.bcast.Robj2slave(furfineData_filename)
	mpi.bcast.Robj2slave(jobname)
	mpi.bcast.Robj2slave(Simulation)
	mpi.bcast.Robj2slave(numboots)
	mpi.bcast.cmd(print("Initialized slaves and their logs\n"))
	print(paste('methodTreatDuplicates = ', methodTreatDuplicates))

	mpi.bcast.Robj2slave(furfdat)
	mpi.bcast.Robj2slave(nMatSlaves)
	mpi.bcast.Robj2slave(FurfineScript)
	mpi.bcast.Robj2slave(bpsh)
	mpi.bcast.Robj2slave(bpsl)
	mpi.bcast.Robj2slave(refinement)

	if(length(args) > 10){
	    print('additional args used for Rempel script \n')
	    mpi.bcast.Robj2slave(SyntheticTraining)
	    mpi.bcast.Robj2slave(useReturnLegCharacteristics)
	    mpi.bcast.Robj2slave(ControlSyntheticForParticipants)
	    mpi.bcast.Robj2slave(useDANN)
	    mpi.bcast.Robj2slave(klossWeighted)
	    mpi.bcast.Robj2slave(Xpaythreshold)
	    mpi.bcast.Robj2slave(ModelTraining)
	    mpi.bcast.Robj2slave(methodDup2)
	    mpi.bcast.Robj2slave(choiceK)
	    mpi.bcast.Robj2slave(subsample_number)
	}

	mpi.bcast.Robj2slave(nMatSlaves)
	mpi.bcast.Robj2slave(Parallel)
	mpi.bcast.Robj2slave(Nchunks)
	mpi.bcast.Robj2slave(methodTreatDuplicates)
	mpi.bcast.Robj2slave(min_sentime)
	mpi.bcast.Robj2slave(min_loansize)
	mpi.bcast.Robj2slave(outputfilename)


	# Setting parameters in the matlab worker workspaces:

	mpi.bcast.cmd(print('Passing parameters to mat workers'))

	mpi.bcast.cmd(evaluate(matlab, "fprintf('[%d] Getting parameters', rank);"))

	mpi.bcast.cmd(setVariable(matlab, Nchunks=Nchunks,
		Parallel=Parallel, nMatSlaves=nMatSlaves, min_sentime=min_sentime,
		min_loansize=min_loansize, methodTreatDuplicates=methodTreatDuplicates,
		refinement=refinement, 
		output_string = outputfilename
	))
	if(length(args) > 10){
	    mpi.bcast.cmd(setVariable(matlab, Xpaythreshold=Xpaythreshold,
		    choiceK=choiceK,
		    ControlSyntheticForParticipants=ControlSyntheticForParticipants,
		    useReturnLegCharacteristics=useReturnLegCharacteristics,
		    useDANN=useDANN,
		    klossWeighted=klossWeighted,
		    ModelTraining=ModelTraining,
		    Simulation=Simulation,
		    methodDup2=methodDup2,
		    subsample_number=subsample_number
	    ))
	}

	mpi.bcast.cmd(evaluate(matlab,
        "fprintf('[%d] number of matlab slave: %d', rank, nMatSlaves);"))
	mpi.bcast.cmd(evaluate(matlab,
        "fprintf('[%d] parameters uploaded', rank);"))

	## Starting

	# adding additional matlab libraries
	mpi.bcast.cmd(evaluate(matlab,
		"addpath('KNN');"))

	# set seed for each matlab worker?

	mpi.bcast.cmd(evaluate(matlab,
		"fprintf('[%d] subsample number: %d', rank, subsample_number);"))

	mpi.bcast.cmd(print("Main Function About To Start"))

	# tags sent by the master to the slaves
	master.tag.work <- 1
	master.tag.done <- 2
	# tags sent by the slaves to the master
	slave.tag.ready <- 101
	slave.tag.finished <- 102
	slave.tag.quit <- 103

	mpi.bcast.cmd({
		master.tag.work <- 1;
		master.tag.done <- 2;
		slave.tag.ready <- 101;
		slave.tag.finished <- 102;
		slave.tag.quit <- 103
	})

	mpi.bcast.Robj2slave(slavefunction)
	mpi.bcast.cmd(slavefunction())

	# Master Function

	numboots <- length(BootSample)
	tasks <- vector('list', length=numboots)
	tasks <- lapply(1:numboots, function(x) list(bootNumber=x, bootsample=BootSample[[x]]$bootsample))
	Results <- vector('list', length=numboots)
	junk <- 0
	closed_slaves <- 0
	n_slaves <- mpi.comm.size() - 1
	while(closed_slaves < n_slaves){
	# receive any message from any slave
		proc.time()[3] -> master.initial.time
		slave.msg <- mpi.recv.Robj(mpi.any.source(), mpi.any.tag())
		slave.msg.info <- mpi.get.sourcetag()
		slave_id <- slave.msg.info[1]
		slave_tag <- slave.msg.info[2]
		print(paste('slave id', slave_id, ' slave_tag ', slave_tag))
		if (slave_tag==slave.tag.ready) {
			if(length(tasks) > 0){
				this.task <- tasks[[1]]
				tasks[1] <- NULL
				print(paste('Number of tasks left -', length(tasks)))
				mpi.send.Robj(this.task, slave_id, tag=master.tag.work);
			} else mpi.send.Robj(obj=NULL, dest=slave_id, tag=master.tag.done)
		} else if (slave_tag==slave.tag.finished) {
		# then we're receiving results
			print(paste('Received Results', 'Subbing them into Results', sep='\n'))
			Results[[slave.msg$bootNumber]] <- slave.msg
		} else if (slave_tag==slave.tag.quit) {
		# A slave has shut down
			print('Slave has closed down')
			closed_slaves <- closed_slaves + 1
			if (n_slaves - closed_slaves==1){
				proc.time()[3] -> master.last.slave.time
			}
		}
		flush.console()
		if(n_slaves - closed_slaves==1){
			if(master.last.slave.time + 120 < proc.time()[3]){
				print('Master has hung-up on the last task')
				break
			}
		}
	}
	cat("Job has run it's course. Exiting cluster\n")
	cat("closing matlabs\n")
	mpi.bcast.cmd(close(matlab))
	mpi.close.Rslaves()
	mpi.quit()
}
