#########################################################################
# 		Master R script - Bootstrap Sampling Algorithm
#				Mark Rempel
#			
# Usage: bsub_RMPI 2 BootstrapBalancedSampling.R --args residualLVTSdatafilename, outputfilename, number of script iterations,  number of return legs per send leg, number of send legs per bootsample, number of bootsamples
#
# This script generates (or takes as given) the set of payments and samples/pairs non-adjacent days together
#
# For a given sampled sender leg day, there will be n return days sampled to be paired with it
#
# Inputs: 1) RData file of payments data contained in object rlvts, which is payments data but with the dates converted from numeric to date format (e.g. as.Date(rlvts$date, format='....'))
#         the rest are arguments with sufficient description given below
#
# Output: RData file(s) with object BootSample, (numBootSamples = arg[3] specifies the number of RData files/repetitions of the sampling to do)
#           length(BootSample) = numboots = B^N per RData file
#           BootSample[[i]] is a list with two elements:
#                 BootSample[[i]]$bootsample is the payments data, with synthetic date labels (in numeric format) such that adjacent dates are payment days to be considered as lend/repayment days
#                 BootSample[[i]]$SRdates is a n_return_size x 2 matrix of the true dates for the re-paired lend / repayment dates
#   
# 			May, 2014
#########################################################################

# Inputs and Global Parameters
args <- commandArgs(trailingOnly = TRUE) # take the arguments trailing the Rscript in batch submission as inputs

# e.g. bsub numcores Rscript.R 'inputfilename' numboots
outputfilename <- args[2] # output file name, no extension (pasting bootsize/etc to the end of outptufilename
residualLVTSdata <- args[1] # the input payments data for the block bootsample 
numBootSamples <- as.numeric(args[3]) # numeric integer of number of .RData file samples of numboots number ofsamples
n_returnsize <- as.numeric(args[4]) # is the number of repayment days per boostrap sample
bsamplesize <- as.numeric(args[5]) # is the number of lend days per bootstrap sample
numboots <- as.numeric(args[6]) # is the number of bootstrap samples (i.e. number of bsamplesize * n_return size sets per RData file)
Seed <- args[7]
print(paste('num RData files to be outputted ', numBootSamples))
print(paste('number of Return days per lend day ', n_returnsize))
print(paste('number of lend days' , bsamplesize))
print(paste('Seed is ', Seed))
nMatSlaves <- 0
set.seed(Seed)


block_sample <- function(X, dates, n_lend, n_returns, nonLendDates){
		# sample size n (distinct) days lent
		# 	for a given date, sample n_returns repayment dates to be paired with it	
		#	and convert the labelling of dates s.t. the synthetic repayment dates are the next day to the lent date 
		# nonLendDates are the dates that can be sampled as repayment days but not lend days (if any)
		# yielding a blocksample of n*n_returns days

		# the output is a list: a dataframe of blocksample of residual payment dates, a dataframe of the true lend and repay dates 

                lend_days <- sample(dates[which(!dates %in% nonLendDates)], replace=T, size=n_lend)
		# sampling from all dates which share the same ON rate and are not the same or + 1 to the lend date
		SameONrateDates.l <- lapply(lend_days, function(x) unique(X[which(X[,7]== unique(X[which(x==X[,1]),7]) & !X[,1]==x & !X[,1]==(x + 1)),1]))
		repay_days <- lapply(SameONrateDates.l, function(x) sample(dates[which(dates %in% x)], replace=T, size=n_returns))

		SRdates.l <- lapply(1:length(lend_days), function(x) data.frame(rep(lend_days[x], n_returns), repay_days[[x]]))
		SRdates <- do.call(rbind, SRdates.l)
		total_size <- n_lend * n_returns
		ldates_gen <- seq(as.Date('2000-01-01'), by='4 days', length=total_size) # generate a lend date to be adjacent to return date
					# implying generate a new lend date label for each (lend date, return date) pair

		ReturnDaysWithLendDays.l <- lapply(1:length(lend_days), function(x){
			# generating the payments and relabelled dates for each pair
			Lend.day <- X[which(X$date==lend_days[x]),]
			# run through each return day for the lend day 
			Repay.days.relabelled <- lapply(1:n_returns, function(repay_date_indx){
				repay_date <- repay_days[[x]][repay_date_indx]
				Return.day <- X[which(X$date==repay_date),]
				rdate_gen <- ldates_gen[(x-1)*20 + repay_date_indx] + 1
				Return.day$date <- as.numeric(gsub('-', '', rep(rdate_gen, nrow(Return.day))))
				return(as.matrix(Return.day))
			})
			# duplicate the lend day n_return times and relabel each one to ldates_gen

			Lend.days.relabelled <- lapply(1:n_returns, function(repay_date_indx){
				ldate_gen <- ldates_gen[(x-1)*20 + repay_date_indx]
				Lend.day$date <- as.numeric(gsub('-', '', rep(ldate_gen, nrow(Lend.day))))
				return(as.matrix(Lend.day))
			})
			repay.days.final <- do.call(rbind, Repay.days.relabelled)
			lend.days.final <- do.call(rbind, Lend.days.relabelled)
			RePairedReturnDaysWithLendDay <- rbind(lend.days.final, repay.days.final)
			RePairedReturnDaysWithLendDay <- RePairedReturnDaysWithLendDay[order(RePairedReturnDaysWithLendDay[,1]),] 
			return(RePairedReturnDaysWithLendDay)

		})

		Day_pairs <- do.call(rbind, ReturnDaysWithLendDays.l)


                ## make data into the form of the furfine algorithm

		blockbootsample <- list(bootsample = Day_pairs, SRdates = SRdates)
		return(blockbootsample)
}
#if(!is.null(num_cores)) mpi.bcast.Robj2slave(block_sample)

print(paste("Creating New Bootstrap Sample File"))
print("Loading payments data")
load(residualLVTSdata) 
print(rlvts$date[1])
dates <- unique(rlvts$date)
NonLendDays <- dates[c(length(dates))] # the last payment cannot be a lend day

# Acquiring numBootSamples of re-paired sets of payment days, each set has bsamplesize amount of lend days and n_returnsize of return days paired to a given lend day
	# i.e. numBootSamples = 2, bsamplesize = 10, n_returnsize = 100 implies
		# for each BootSample[i] (i = 1 or 2)
		# 	there will be a data.frame with 10 sampled lend dates (likely but not necessarily unique)
		#		for each of the 10 lend dates, there will be 100 return dates paired with it
		#		yielding a total of 1,000 lend/return date pairs (in a given BootSample i)
limit <- 9*1073741824 # 6.6 Gb
for(i in 1:numBootSamples){
	print(paste('Starting', i))
	BootSample.l <- lapply(1:numboots, function(x) block_sample(rlvts, dates, bsamplesize, n_returnsize, NonLendDays))
	BootSample <- BootSample.l
	if(object.size(BootSample) > limit){
		# split into two pieces
		if(numboots==1){
			BootSample <- list(BootSample.l[[1]]$bootsample[1:floor(nrow(BootSample.l[[1]]$bootsample)/2),], BootSample.l[[1]]$SRdates)
			save(BootSample, file=paste(outputfilename, '_NumLendDays_', bsamplesize, 'NumReturnDays_', n_returnsize, '_Numboots_', numboots, '_', i,'A.RData', sep=''))
		        BootSample <- list(BootSample.l[[1]]$bootsample[(1 + floor(nrow(BootSample.l[[1]]$bootsample)/2)):nrow(BootSample.l[[1]]$bootsample),], BootSample.l[[1]]$SRdates)
	                save(BootSample, file=paste(outputfilename, '_NumLendDays_', bsamplesize, 'NumReturnDays_', n_returnsize, '_Numboots_', numboots, '_', i,'B.RData', sep=''))
		} else {
                        BootSample <- lapply(1:floor(numboots/2), function(x) BootSample.l[[x]])
                        save(BootSample, file=paste(outputfilename, '_NumLendDays_', bsamplesize, 'NumReturnDays_', n_returnsize, '_Numboots_', numboots, '_', i,'A.RData', sep=''))
			BootSample <- lapply((floor(numboots/2)+1):numboots, function(x) BootSample.l[[x]])
                        save(BootSample, file=paste(outputfilename, '_NumLendDays_', bsamplesize, 'NumReturnDays_', n_returnsize, '_Numboots_', numboots, '_', i,'B.RData', sep=''))


		}

	} else {
		save(BootSample, file=paste(outputfilename, '_NumLendDays_', bsamplesize, 'NumReturnDays_', n_returnsize, '_Numboots_', numboots, '_', i,'.RData', sep=''))
	}
}


print('All Done')

