%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Control Script of Furfine Algorithm and K-NN Variants
%
% Overview:
%	Script is implementation of Furfine algorithm (1999), Hendry et al. (2008) and k-NN variants
%		as discussed in Improving Overnight Loan Identification through Payment Systems (Rempel, 2014)
%
%	Script should be called from original_furfine_jobs file (bsub -pack commands)
%
%	The script requires the specification of numerous parameters/inputs as specified therein.
%		e.g. upper/lower bounds on the interest rate thresholds, minimum loan principal size, etc.
%	
%	The script allows for both parallel and non-parallel implementations
%
% Inputs (Needed in workspace):
%	lvts = payments data in numeric matrix format or cell of numeric matricies
%	Nchunks = number of (desired) partitions of the data		
%		if Nchunks > 1 and lvts is numeric matrix, then partitioned data is created
%	Parallel = indicator (1 - use parallel methods, 0 otherwise)
%		if Parallel = 1 need:
%			nMatSlaves = number of matlab slaves to be requested by matlabpool open
%		
%
%	Simulation = indicator of whether inputted payments data is to be treated as original (0), or from resampled dates (Simulation = 1)
%
%	Furfine Algorithm parameters:
%		min_sentime = minimum sendtime of payments to be considered (e.g. 6 a.m is start of normal payments)
%		bpsh/bpsl = upper/lower interest rate threshold bound from target (e.g 0.5 is 50 basis points above target)
%		min_loansize = minimum payment size to be considered for overnight loans
%		
%	MethodTreatDuplicates = {'Hendry', 'Armantier', 'Rempel'}
%		'Hendry' is closest to interest rate duplicate potential loan criterion
%		'Armantier' is to take the payment pair with the median interest rate from the set of duplicates
%		'Rempel' is the k-NN criterion described in Rempel, 2014
%
%		If Rempel selected:
%			ControlSyntheticForParticipants = 1 (alternative of 0 is where synthetic `faux' loans paired without consideration of the sender/receiver)
%			useReturnLegCharacteristics = 1 (alternative of 0 indicates only payment characteristics of the initial send payment should be used)
%				Note: setting either of the two to 0 is not necessarily supported in the scripts at the current time
%			methodDup2 = {'Hendry', 'Armantier'}
%				Indicates the criterion to be used in the case that the k-NN criterion fails to identify a unique payment pair from the duplicate loan sets
%		
%	
%	Note: the path to supplemental functions must also be specified 
%		
% Outputs:
%	Loans = identified overnight loans by the selected algorithm/variant
%	LOANS = the identified overnight loans by each partition 
%	
% Last Modified:
%	Rempel, Feb. 2014
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


disp('Initiating Overnight Loan Identification')
if ~strcmp(methodTreatDuplicates, 'Rempel')==1
        ModelTrainingData = 0;
end

sprintf(['Running with refinement ' refinement '\n'])

%% Canada specific conventions

ond = 365; % Actual/365 interest rate computation


%% Payments data format


format long g
% symmetric bounds

% specifying column locations for variables of interest

datecol= 1;
timecol = 2;
sdate_column = 1;
rdate_column = 11;
senderValue_column = 3;
returnValue_column = 13;
sender_column = 4;
receiver_column = 5;
intrate_column = 21;
ONrate_column = 7;
senderTransactionIdColumn = 10;

SenderDuplicateIndicatorValue = 4;
SenderSelectedDuplicateValue = 2;
receiverTransactionIdColumn = 20;
ReceiverDuplicateIndicatorValue = 6;
ReceiverSelectedDuplicateValue = 3;


if ~exist('endofdaythreshold', 'var')
	endofdaythreshold = min_sentime; % setting end of day threshold if missing to payment system start time 
end
%%%% End of Parameterizations %%%%

%% Begining pre-processing of inputs
	% if Nchunks > 1 partitions the payments data into that number of continguius cells to be processed in parallel or reduce memory burden

if isa(lvts, 'double') && Nchunks > 1

    if(length(unique(lvts(:,1)))< 2)
            disp('Less than 2 unique dates found in data')
            disp('Aborting Algorithm')
    end

    if size(lvts,2) < 10
        tnum = 1:size(lvts,1);
        lvts = [lvts, tnum'];
    end

    LVTS = sortrows(lvts,1:2);
    LVTS = LVTS(LVTS(:,1) > 0,:);
    datecol= 1;
    clear lvts

    disp(['number of unique dates = ' num2str(length(unique(LVTS(:,1))))])

    % run Partitioning function
	partitionFilename = ['Inputs/LVTS_Partitions' num2str(Nchunks) '.mat'];
	if exist(partitionFilename, 'file') == 2
		sprintf(['Partitioned Payments data found for specified Nchunks. Loading file ' partitionFilename])
		clear lvts
		load(partitionFilename)
	else
	              sprintf(['Partitioned Payments data found for specified Nchunks. Loading file ' partitionFilename])
   
	    Partitions = ContiguousPartitions(LVTS, datecol, Nchunks);
	    eval(['save(''Inputs/LVTS_Partitions' num2str(Nchunks) '.mat'', ''Partitions'', ''-v7.3'')']) 
	end
else
    if Nchunks == 1
        Partitions = cell(Nchunks,1);
        Partitions{1} = lvts;
    else
        Partitions = lvts;
    end
    
end

% resizing Nchunks to the actual number of partitions
if iscell(Partitions)
    Nchunks = length(Partitions); 
else 
    Nchunks = 1;
end
LOANS = cell(Nchunks,1);


%% Applying the specified overnight loan identification algorithm
if Parallel==1
    %% Parallel version
    parfor partitionI = 1:Nchunks
        % setting random stream
        sprintf('partitionI now %d (in parallel)\n', partitionI)
        randomstream = RandStream.getGlobalStream;
        randomstream.Substream = partitionI;
        % run furfine function on a given partition
        partitionIdata = Partitions{partitionI};
        FurfineData = BaselineFurfine(partitionIdata, min_sentime, min_loansize, bpsh, bpsl, ond,methodTreatDuplicates, refinement,Simulation, endofdaythreshold);

        if strcmp(methodTreatDuplicates, 'Rempel')
             %% KNN Training and Prediction
	    fprintf('Running Rempel Refinement \n')

            run RempelRefinement; 

            LOANS{partitionI} = [FurfineData(FurfineData(:,22)==0,:); PredData];    

        else
            %% Non-Rempel assigment

            LOANS{partitionI} = FurfineData; % the censoring of dup values in column 22 has already been done (in the Hendry/Armantier case)

        end
    end
else
    %% Serial Version
    
    for partitionI = 1:Nchunks
        % run furfine function on a given partition
        
        sprintf('partitionI now %d (in serial) \n', partitionI)

        partitionIdata = Partitions{partitionI};

        FurfineData = BaselineFurfine(partitionIdata, min_sentime, min_loansize, bpsh, bpsl, ond, methodTreatDuplicates, refinement,Simulation, endofdaythreshold);

	fprintf('Baseline Furfine complete \n')

        if strcmp(methodTreatDuplicates, 'Rempel')
            %% KNN Training and Prediction
       
	    fprintf('Running Rempel Refinement \n')
    
            run RempelRefinement; 
	 
            LOANS{partitionI} = [FurfineData(FurfineData(:,22)==0,:); PredData];
        else
            %% Non Rempel Assignment

             LOANS{partitionI} = FurfineData; % the censoring of dup values in column 22 has already been done (in the Hendry/Armantier case)
            
        end
    end
end

Loans = cell2mat(LOANS);
Loans = unique(Loans, 'rows');
disp('Loan identification complete')





