function FurfineData = BaselineFurfine(data, minimumSendTime, minimumLoanSize, bpsh, bpsl, ond, duplicateTreatmentMethod, refinement, IsSimulation, varargin)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Baseline Furfine algorithm
%
% Standard Furfine (1999) algorithm with the duplicate treatment
% criterion of Hendry et al. (2007) or Armantier/Copeland (2012), or 
% Rempel (2013). See any of the above for details on the 
% alogrithm
%
% Takes payment data (date, time, value, sender, receiver, ONrate, transaction identifier, ..)
%   and creates a plausible loan and repayment dataset 
%   by matching interday payment data together that seem to mimic
%   overnight loans to each other
%
% Inputs:	
%	(partitions) of contiguous days of payments data, with minimum time, loan size thresholds, interest rate upper/lower bounds, duplicate treatment method 
%		as well as an indicator of whether or not the dates are to be taken literally or not (IsSimulation)
%
% Outputs: 
%	FurfineData - n x 22 matrix of potential loans indicated by the baseline Furfine algorithm (1999)
%		columns 1 - 10 are the payment characteristics of the loan payment
%		columnes 11 - 20 are the payment characteristics of the loan repayment 
%		column 21 is the implied interest rate computed on an actual/365 basis (unless ond modified)
%		column 22 indicates whether or not the payments involved are unique matches (value of 0) or 	
%
% Written
% 	Rempel,  2013, modified Feb. 2014
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

     %disp(['size within nonRempelScript ' num2str(size(St1,1))])

    if length(varargin) > 0
	endofdaythreshold = varargin{1};
    end

    % Payment data matrix columns
	BoCid = 2;
	datecol= 1;
	timecol = 2;
	sdate_column = 1;
	rdate_column = 11;
	senderValue_column = 3;
	returnValue_column = 13;
	sender_column = 4;
	receiver_column = 5;
	intrate_column = 21;
	ONrate_column = 7;
	senderTransactionIdColumn = 10;

   % Duplicate Indicator Codes
	SenderDuplicateIndicatorValue = 4;
	SenderSelectedDuplicateValue = 2;
	receiverTransactionIdColumn = 20;
	ReceiverDuplicateIndicatorValue = 6;
	ReceiverSelectedDuplicateValue = 3;

    dates = unique(data(:,datecol));
    firstdate = min(dates);
   
    % imposing minimum size restriction
    data = data(data(:,senderValue_column) >=minimumLoanSize,:);
    data = data(data(:,timecol) >= minimumSendTime,:); % eliminating all loans prior to..

    % For FAD dates (days in which the ON rate is announced change, the
    % effective interest rate is the prior days until 9am,, where upon the
    % newly announced one takes effect)

    R = [zeros(size(find(data(:,datecol)==firstdate),1), size(data,2)); data(data(:,datecol)>firstdate,:)]; % [0_nxm; lvts_day2+]
    

    % Restricting sample
    f = find(data(:,senderValue_column)-round(data(:,senderValue_column))==0);   % Round dollar-value originating loans.
    data = data(f,:); %#ok<FNDSB>

%    disp(['lvts size within nonRempelScript ' num2str(size(data,1))])
    Dates = datenum(num2str(dates), 'yyyymmdd');
    lagDates = lagmatrix(Dates,1);
    lagDates(1) = NaN;
    maxT = max(Dates - lagDates);
    if IsSimulation==1
            % the days are serially uncorrelated, we only wish to test pair-wise matching
            disp('Simulation==1, therefore max interval between days is 1 (maxT)')
            maxT = 1;
    end

    if maxT > 7
        % no overnight loans possible in 7 day horizon ..
        maxT = 7;
    end
    
    
    furfineData = cell(maxT, 1);
    

    for t = 1:maxT

        
        % Receiver matrix 
        r1 = min(dates(dates>firstdate));
        if size(r1,1)==0
                disp('Nchunks is too large, there are not multiple dates in a given chunk')
                break
        end
        
        Rt1 = R(R(:,datecol)>= r1,:);

        % Selecting only the dates which have a difference of t-days
        % between consecutive dates (e.g. monday - friday = 3 days)
        
        t_interval_days = find(Dates - lagmatrix(Dates, 1)==t) - 1;

        next_tday = t_interval_days + 1;
        

        if size(t_interval_days, 1) <= 0
                %fprintf('Skipping %d interval days \n', t)
                continue
        end

        %fprintf('Running %d interval day iteration \n', t)


        % t_Sinterval_dates are the Sender dates with only t_intervals between
        t_Sinterval_dates = intersect(dates(t_interval_days), data(:,datecol));
        t_Sinterval_idx=find(ismember(data(:,1), t_Sinterval_dates));
        St1 = data(t_Sinterval_idx,:); %#ok<FNDSB>

        % dates(t_interval_holidays + 1) = earliest possible repayment

        t_Rinterval_dates = intersect(dates(next_tday), Rt1(:,1));
        t_Rinterval_idx = find(ismember(Rt1(:,1), t_Rinterval_dates));
        Rt1 = Rt1(t_Rinterval_idx,:); %#ok<FNDSB>
        
        %  Value_t \in  [(1 + %(daily int rate +/- bps)) *Value_(t+m) *m]
        I = [];
        for i = 1:size(St1,1)
                s1d = datenum(num2str(St1(i,datecol)), 'yyyymmdd');
                s1dPlus1 = s1d + t;
                s1dPlus1n = str2double(datestr(s1dPlus1, 'yyyymmdd'));
                ConsecutiveDays = find(Rt1(:,1)==s1dPlus1n);
                    
                SRpairs = find(St1(i,receiver_column)==Rt1(:,sender_column) & St1(i, sender_column)==Rt1(:,receiver_column));
                SendReceiveMatches = intersect(ConsecutiveDays, SRpairs);
                InterestWithinThreshold = find(Rt1(:,senderValue_column) <= (((((St1(i,ONrate_column)+bpsh)./(100*ond)).*t)+1).*St1(i,senderValue_column)) & (((((St1(i,ONrate_column)-bpsl)./(100*ond)).*t)+1).*St1(i,senderValue_column)) <= Rt1(:,senderValue_column));
                PotentialLoans = intersect(SendReceiveMatches, InterestWithinThreshold);
                if size(PotentialLoans,1) > 0
                    if size(I,1) > 0
                        % aggregating the indices of the potential
                        % loans
                        % I contains the rows of data
                        % J contains the corresponding row in R
                        J = [J; PotentialLoans]; 
                        is = zeros(length(PotentialLoans),1) + i;
                        I = [I; is]; %#ok<AGROW>
                        clear is;
                    else
                        % first set of potential loans
                        I = zeros(length(PotentialLoans),1) + i;
                        J = PotentialLoans;
                    end
                end
        end
        %disp('Potential Loans Populated')

        %   creating a matrix loans =[PotentialLoanGiven_t,   PotentialRepayment_t+1, FurRate]


        NumberOfPotentialLoans = size(I,1);
        if NumberOfPotentialLoans <= 0
            %fprintf('Number of Potential Loans less than or equal to 0, moving to next %d t iteraion', t)
            continue
        else
            %disp('Number of potential loans is positive')
        end
        SRindices = [I,J];
        Z1o = [St1(SRindices(:,1),:), Rt1(SRindices(:,2),:)];

        % furf_rate = annual interest rate in basis points
        furf_rate = (((((Z1o(:,returnValue_column)./Z1o(:,senderValue_column))-1).*(100)).*ond)./t);
        Z1a = [Z1o, furf_rate];
        loans = unique(Z1a, 'rows');
        
        % removing any transactions involving the Bank of Canada
        loans(loans(:, sender_column)==BoCid | loans(:, receiver_column)==BoCid,:) = [];

        % find all positive furfine rates
        loans = loans(loans(:,intrate_column) > 0, :);
        loans = sortrows(loans,[1 4 5]); % sort by sender date, sender participant, receiver participant
        
        loans = [loans, zeros(size(loans,1),1)]; %#ok<AGROW> % making an indicator variable for duplicate loans
        
        % Selecting the smaller of the rates in the case of same originating
        % payment or return payment
            % assuming banks will tend to offer rates closer to the target
     	
	%% Apply any pre-duplicate treatment to potential overnight loans

	if strncmp(refinement, 'IntRound', 8)
		% apply rounded interest rate criterion, ensuring exact repayment value
	
		sdate_str = arrayfun(@(x) num2str(x), loans(:, sdate_column), 'UniformOutput', false);
		rdate_str =  arrayfun(@(x) num2str(x), loans(:, rdate_column), 'UniformOutput', false);
		sdate_num = datenum(sdate_str, 'yyyymmdd');
		rdate_num = datenum(rdate_str, 'yyyymmdd');
		day_diff = rdate_num - sdate_num;
		if strcmp(refinement, 'IntRoundExact')
			IntRoundExactIndicator = InterestIsRounded(loans(:, senderValue_column), loans(:, returnValue_column), loans(:, intrate_column), day_diff);
			loans(IntRoundExactIndicator==0,:) = [];
			percentremoved = length(find(IntRoundExactIndicator==0))/length(IntRoundExactIndicator);
			%sprintf('Applied Exact Rounded Interest Rate criterion. Removed %d percent of potential loans.', percentremoved)
		else
			 % apply rounded interest rate criterion, (here only checking that rounded to the nearest tenth of a bp, interest rate is 0 or 5)
                        	% i.e. after rounding, interest rate is quoted at whole or half basis points
                        [~, IntRound] = InterestIsRounded(loans(:, senderValue_column), loans(:, returnValue_column), loans(:, intrate_column), day_diff);
			loans(IntRound==0,:) = [];
                        percentremoved = length(find(IntRound==0))/length(IntRound)*100;
                        %sprintf('Applied Rounded Interest Rate criterion. Removed %d percent of potential loans.', percentremoved)
  
		end
	end
	if strcmp(refinement, 'EndOfDay')
		percentremoved = length(find(loans(:, timecol) < endofdaythreshold))/length(loans(:,1))*100;
		loans(loans(:, timecol) < endofdaythreshold,:) = [];
                %sprintf('Applied End of Day criterion, with threshold %d. Removed %d percent of potential loans.', [endofdaythreshold,percentremoved])
	end

	%% Duplicate loan treatment

        loans = BaselineDuplicateTreatment(loans, senderTransactionIdColumn, ONrate_column, intrate_column, duplicateTreatmentMethod, SenderDuplicateIndicatorValue, SenderSelectedDuplicateValue);
        
        loansNonUniqueAndNotSelected = loans(loans(:,22) > SenderSelectedDuplicateValue,:);
        loans = loans(loans(:,22) <= SenderSelectedDuplicateValue,:);
    
        % Receiver payment duplication (same as sender)
        
        loans = BaselineDuplicateTreatment(loans, receiverTransactionIdColumn, ONrate_column, intrate_column, duplicateTreatmentMethod, ReceiverDuplicateIndicatorValue, ReceiverSelectedDuplicateValue);
        % given SenderSelectedDuplicateValue = 2, SenderDuplicateValue = 4
        %       ReceiverSelected.. = 3, ReceiverDuplicate.. = 6
        
        %  the joint indicator is: (addition of two indicators excluding
        %  SenderDuplicateValue which we prevent the second round duplicate check from
        %  running on
        
        % loan duplicate indicator values (column 22) {0 = unique; 2 =
        % sender selected, receiver unique; 3 = sender unique, receiver
        % selected; 5 = sender/receiver both selected; 6 = sender unique,
        % receiver not selected; 8 = sender selected, receiver not
        % selected}
        
        if strcmp(duplicateTreatmentMethod, 'Rempel')
            % combine the loansNonUniqueAndNotSelected and those selected
            %   as we are using a different duplicate criterion
            loans = [loans; loansNonUniqueAndNotSelected]; %#ok<AGROW>
            loans = sortrows(loans, [1 2]);
            furfineData{t} = loans;
        else
            % keeping only the potential loans which were unique or
            % selected using Armantier/Hendry criterion from duplicates
            furfineData{t} = loans(loans(:,size(loans,2))<= SenderSelectedDuplicateValue + ReceiverSelectedDuplicateValue,:); % <= 5 with default sender/receiverSelected values
        end
    end
    FurfineData = cell2mat(furfineData);
    if size(FurfineData,1) > 0 
        a = datenum(num2str(FurfineData(:,11)), 'yyyymmdd');
        b = datenum(num2str(FurfineData(:,1)), 'yyyymmdd');
        if max(a - b) > 7
            disp('Error, exceeded reasonable length for overnight loan')
            save('ErrorFurfine.mat', 'FurfineData')
            return
        end
    
    else
       disp('No loans identified in this sample'); 
    end
end

