function Partitions = ContiguousPartitions(data, datasplit_column, N)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Pre-processing function, supplemental to FurfineControlScript.m
%	Simply splits payments data into N separable continguous chunks (wrt payment days)
%		Intended for parallelized running of Furfine algorithm/kNN variants (as described in Rempel (2014)
%
%	Partitions the data, keeping all payments of a given day together
%		with a minimum of 2 adjacent payment days together in a partition (NOTE ideal partition size incorporates at least a week-to-month of data)
%		
%		Last partition will combine the days which were at the edges of the partitions (i.e. last day of partition 1 and first day of partition 2 combined)
%		
%
%	NOTE: this script must be substantially modified for use of loans greater than overnight in maturity
%
%	Inputs:
%		Payments data
%		column of payments data which contains the dates
%		N = the number of partitions to be created
%
%	Outputs:
%		Payments data in N partitions (a cell, with each individual cell a subset of the payments data)
%
% Rempel, 2013
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    %  take the number of unique days, split that into N chunks, with one
    %  date overlap
    
    unique_values = unique(data(:, datasplit_column));
    
    % split these unique_values into N contiguous chunks 
    
    
    num_values = length(unique_values);
    edges = 1:ceil(num_values/N):num_values;
    if edges(length(edges)) < num_values 
        edges = [edges num_values];
    end
    %partitions = cell(N, 1);
    idx = arrayfun(@(x) edges(x):edges(x+1), 1:length(edges)-1, 'UniformOutput', false);
    partitions = cellfun(@(idxx) cell2mat(horzcat(arrayfun(@(oneindxx) data(unique_values(oneindxx)==data(:, datasplit_column),:), idxx, 'UniformOutput', false)')),idx, 'UniformOutput', false);
    Partitions = cell(length(partitions) + 1, 1);
    Partitions(1:length(partitions)) = partitions;
    % making the Partitions have one day of overlap to get full coverage
    headtailjoinIndx = cell2mat(arrayfun(@(x) edges(x)-1:edges(x), 2:length(edges), 'UniformOutput', false));
    a = cell(length(headtailjoinIndx),1);
    for kk = 1:length(headtailjoinIndx)
        oneindxx = headtailjoinIndx(kk);
        a{kk} = data(unique_values(oneindxx)==data(:,datasplit_column),:);
    end
    Partitions{end} = cell2mat(horzcat(a));
    
end
    
    
        
    
