function Covariates = GeneratingKNNcovariates(Xdata,varargin)
% generate predictor covariates 
%   create the independent variables used for classification of duplicate loans
%
% Inputs:
%	Xdata - raw loan data
%
% Outputs:
%	Covariates - summary variables of loan data
%       	1) duration (number of hours between payment sent and repaid, standardized)
%		2) rounded interest rate indicator 
%       	3) actual repayment value - implied repayment value of loan based on interest rate rounded to nearest basis point (standardized)
%		4) send payment time (in hours from midnight, standardized)
%       	5) interest rate spread from target (standardized)
%		6) principal payment value (standardized)
% Rempel, 2013

    nargs = length(varargin);
    if nargs > 0 
        UseBestPredictorsOnly = varargin{1};
    else
        UseBestPredictorsOnly = 0;
    end
    
    stime_column = 2;
    rtime_column = 12;
    rdate_column = 11;
    sdate_column = 1;
    ONrate_column = 7;
    int_rate_column = 21;
    principal_column = 3;
    repayment_column= 13;
    
    send_time = ConvertTimeToElapsed(Xdata(:, stime_column));
    loan_duration = ConvertTimeToElapsed(Xdata(:, rtime_column), Xdata(:, stime_column), Xdata(:, rdate_column), Xdata(:, sdate_column));
    
    principal = Xdata(:, principal_column);
    principal = principal /10^6;
    sdate_str = arrayfun(@(x) num2str(x), Xdata(:, sdate_column), 'UniformOutput', false);
    rdate_str =  arrayfun(@(x) num2str(x), Xdata(:, rdate_column), 'UniformOutput', false);
    sdate_num = datenum(sdate_str, 'yyyymmdd');
    rdate_num = datenum(rdate_str, 'yyyymmdd'); 
    day_diff = rdate_num - sdate_num;
    spread = Xdata(:, int_rate_column) - Xdata(:,ONrate_column);
    [RoundedInterestExact,RoundedInterest, NearestBPdigit, RoundRepaymentDifference]  = InterestIsRounded(principal, Xdata(:, repayment_column), Xdata(:,int_rate_column), day_diff);
    send_time = send_time/3600; %from seconds to hours
    
    loan_duration = loan_duration/3600;
    loan_duration2 = mod(loan_duration, 42); % accounting for weekends etc. 
    %   ensuring the synthetic and the real are directly comparable by normalizing
    
    loan_duration_std = Normalize(loan_duration2);
    principal_std = Normalize(principal);
    send_time_std = Normalize(send_time);
    spread_std = Normalize(spread);
    repaymentdiff_std = Normalize(RoundRepaymentDifference);    
    if UseBestPredictorsOnly==1
        Covariates = [loan_duration_std, SpreadIsRounded, send_time_std];
    else
        Covariates = [loan_duration_std, RoundedInterest, repaymentdiff_std, send_time_std, spread_std, principal_std];

    end
    
    
end
