function TrainedModel = RempelKNNtraining(Xtrain, Ytrain, varargin)
%%% Conducting the Training stage for the K-NN algorithm
%
% Function takes in randomly paired payments (faux-loans) and uniqueley identified loan covariates - Xtrain - 
% 	with binary labels given in Ytrain and outputs a KNNmodel object (using the Mahalanobis metric) 
%
% If no other arguments are supplied, use cross-validation to determine the optimal `K' for K nearest neighbours 
%	If one additional argument, that argument specifies the fixed `K' for K-nearest neighbours
%	If more than one additional argument then the arguments specify options for the selection of optimal k
%		arg1 = klossWeighted =  1 if using binomial deviance as a weighting in considering the choices of K (0 for unweighted variant)
%		arg2 = maxk = specifying the upper limit of the range of K's to check in cross-validation
%			- recommend selecting an odd number > 4
%			will run through the range of values from 1 to maxk (by intervals of 4) when choosing the optimal k
%		
%		arg3 = Parallel = indicator of whether to run cross-validation in serial (0) or parallel (1)
%			- note: if Parallel = 1, assumes a matlabpool has been opened
% Rempel 2013
 
    choiceK = [];
    nargs = length(varargin);
    if nargs==0
	% use defaults for cross-validation
        klossWeighted = 1;
        maxk= 51;
        trialks = 1:4:maxk;
        choiceK = [];
        Parallel = 0;
    else
        if nargs == 1
            choiceK = varargin{1};
        else
            if nargs == 3

                klossWeighted = varargin{1};
                maxk = varargin{2};
                trialks = 1:4:maxk;
                Parallel = varargin{3};
            else
                error('kNNTraining:argIssue', 'Improper number of arguments given')
            end
        end
    end
    if size(Xtrain,1) > 2
        global_cov = nancov(Xtrain);
    else
        global_cov = eye(size(Xtrain,2)); % Identity matrix (euclidean distance)
    end
    
    % now 
 
    if isempty(choiceK)
        CVMCR = NaN(length(1:4:maxk),1);
        if Parallel == 1
            parfor index = 1:length(trialks)
                k = trialks(index);
                 modelk = ClassificationKNN.fit(Xtrain, Ytrain, 'Distance', 'mahalanobis', 'NumNeighbors',k, 'Cov', global_cov);
                  %modelk = ClassificationKNN.fit(Xtrain, Ytrain, 'Distance', 'euclidean', 'NumNeighbors',k);
                  crossvalk = crossval(modelk);
                  if klossWeighted==1
                        kloss =  kfoldLoss(crossvalk, 'lossfun', 'binodeviance'); % use binomial deviance to weight the proportions misclassified
                                                         % is robust too outliers 
                  else
                        kloss = kfoldLoss(crossvalk); % this gives just the raw proportions misclassifed
                  end
                  CVMCR(index) = kloss;
            end
        else
            for index = 1:length(trialks)
                k = trialks(index);
                modelk = ClassificationKNN.fit(Xtrain, Ytrain, 'Distance', 'mahalanobis', 'NumNeighbors',k, 'Cov', global_cov);
                  %modelk = ClassificationKNN.fit(Xtrain, Ytrain, 'Distance', 'euclidean', 'NumNeighbors',k);
                  crossvalk = crossval(modelk);
                  if klossWeighted==1
                        kloss =  kfoldLoss(crossvalk, 'lossfun', 'binodeviance'); % use binomial deviance  
                                                         
                  else
                        kloss = kfoldLoss(crossvalk); % raw proportions misclassifed 
                  end
                  CVMCR(index) = kloss;
            end
        end
        optk = trialks(min(CVMCR)==CVMCR); % selecting k with minimial (weighted)  misclassified proportions
    else
        optk = choiceK; % use user-specified k
    end
    %which ClassificationKNN.fit % just ensuring that the function is there
    TrainedModel = ClassificationKNN.fit(Xtrain, Ytrain, 'Distance', 'mahalanobis', 'Cov', global_cov, 'NumNeighbors', optk);
end
    
    
    
    
    
