function TrainingData = Training_Data(FurfineData, LVTSdatafile, ControlSyntheticForParticipants, varargin)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Supplemental function called from RempelFurfineControlScript
%
%	Function creates a training loan data set which combines the `uniquely' identified loans 
%		with synthetic 'Faux Loans' which are randomly paired/sampled bilateral payments across the entire sample of payments of the days given by FurfineData (payments .mat file = LVTSdatafile)
%		which staisfy the following restrictions:
%       1) The payments must satisfy the same value threshold as the Unique Loans (`True Loans')
%       2) The sampled pair of payments must match the bilateral
%       relationship (i.e. if it's from i - j on day s, then it's j - i on day t
%       3)  Unlike the True Loans, there is no adjacent day restriction, and
%       all paired payments are treated as if they were adjacent days
%       regardless of the true date difference
%	
%	Note: the number of synthetic `Faux' loans will equal the number of unique loans inputted (FurfineData)
%
%	The function will generate synthetic faux loans based on payments between participants (i.e. maintains bilateraly payment relationships)
%
%	Inputs:
%		FurfineData = FurfineData(FurfineData(:,22)==0,:); i.e. the set of uniquely identified loans
%		LVTSdatafile = mat file path of payments data which the Furfine data is a proper subset
%		ControlSyntheticForParticipants = indicator 1 for sampling/pairing payments for synthetic based on bilateral relationships (conditional on participants)
%						alternative of 0 not written
%
%		varargs = alternative furfine restrictions, only the minimum threshold (first arg) accommodated in current function, default is 1 million
%
%	Output: 
%		matrix conformable with Furfine Data
%		output = [UniquelyIdentifiedFurfineData; FauxLoansData]
%			- column 22 of output yields 0 for UniquelyIdentified loans, 1 for faux loans
%
% Written
% 	Rempel 2013
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


    %NOTE: HERE the FurfineData must be s.t. FurfineData =
    %FurfineData(FurfineData(:,22)==0,:); 
    %   that is only the uniquely id'd loans must be in the FurfineData
    disp('in training function')
    size(FurfineData)
    class(LVTSdatafile)
    ControlSyntheticForParticipants
    disp('input values given above')
    nargs = length(varargin);
    if nargs > 0
        % then the Furfine data restrictions are specified:
        min_size = varargin{1};
        min_sentime = varargin{2};
        %min_spread =- varargin{3}; % min absolute spread
    else
        min_size = 1000000;
        min_sentime = 60000;
        
    end
    
    
    datecolumn = 1; % dateindex of LVTS/Furfine data
    datecolumn_return = 11; % column index of furfine data for the return leg
    send_value_column = 3;
    return_value_column = 13;
    sender_id_column = 4;
    receiver_id_column = 5;
    ond = 365; % number of days to annualize overnight rate
    maxFauxSample = 10; % max number of business days to spread out the randomly paired payments into Faux loans

 % Generate training data
           % take all of the uniquely identified sender payments and a subset of the non-id'd sender payments 
                       %  size of subset in [size(uniquely identified sender payments), numpayments in first day]
            
  % step 1 - acquire the corresponding LVTS data and rbind with FurfineData
    if ischar(LVTSdatafile)
        disp('loading LVTSdatafile')
        LVTSpointer = matfile(LVTSdatafile);
        lvts = LVTSpointer.lvts;      
        size(lvts)
        dates = unique(FurfineData(:, datecolumn)); % dates
        dates = sort(dates);
        minFurfineDate = min(dates);
        dates_return = unique(FurfineData(:, datecolumn_return));
        maxFurfineDate = max(dates_return);
        lvts = sortrows(lvts, [1 2]);
        lvts = lvts(lvts(:,1) <=maxFurfineDate & lvts(:,1) >= minFurfineDate,:);
        size(lvts)
        
    else
        lvts = LVTSdatafile; % lvts data already loaded in workspace
        % furfine and lvts data assumed to have already been made similar
        % in time frame (e.g. through the BootstrapModelTrainingScript.R)
    end
    
    LVTSdates = unique(lvts(:,1));
    
        disp('Control Synthetic For Pariticipants is 1')

%% Sample payments from a given pair of participants s.t. the number
        % of pairs per relationship is equal to that in the Furfine Data
        
        RelationshipPairsC = [FurfineData(:,sender_id_column), FurfineData(:, receiver_id_column)];
        
        RelationshipPairs = num2cell(unique(RelationshipPairsC, 'rows'));
        
        index = 1:length(RelationshipPairs);
        Num_ij = arrayfun(@(x) length(find(RelationshipPairsC(:,1)==RelationshipPairs{x,1} & RelationshipPairsC(:, 2)==RelationshipPairs{x,2})), index, 'UniformOutput', false); % number of id'd loan flow from i to j
        
        PaymentPairs = cell(length(RelationshipPairs),1);
        
        % restriction [1] - value > = 1m 
        lvts = lvts(lvts(:, send_value_column)>= min_size,:);
        lvts = lvts(lvts(:, send_value_column)>= min_sentime,:);
        
        for sr = 1:length(RelationshipPairs)
            senderi = RelationshipPairs{sr,1};
            receiveri = RelationshipPairs{sr,2};
            
            SRlvts = lvts(lvts(:, sender_id_column)==senderi & lvts(:,receiver_id_column) ==receiveri,:);
            RSlvts = lvts(lvts(:, sender_id_column)==receiveri & lvts(:,receiver_id_column) ==senderi,:);
            
            
            if size(SRlvts,1) > 0 && size(RSlvts,1) > 0 
                LVTSrelationshipPaymentsSend = SRlvts(randsample(1:size(SRlvts,1), Num_ij{sr}, true),:);
                
                LVTSrelationshipPaymentsReceive = RSlvts(randsample(1:size(RSlvts,1), Num_ij{sr}, true),:);
      
                PaymentPairs{sr} = [LVTSrelationshipPaymentsSend, LVTSrelationshipPaymentsReceive];
            end
            
        end
    
    FauxLoans = vertcat(PaymentPairs{:});
    IntRate = (FauxLoans(:, return_value_column) ./FauxLoans(:, send_value_column) - 1).*(100*ond); % in percent
    FauxLoans = [FauxLoans, IntRate, ones(size(FauxLoans,1),1)]; % Joint payment characteristics, int rate, indicator of false loan
    % replace FauxLoan Return dates, with FauxLoan dates + 1 business day
    maxDays = min(size(LVTSdates,1), maxFauxSample); % take two weeks of business days (or max days in sample) to be assigned to Faux loans
    temp = LVTSdates(1:maxDays);
    tempPos = find(temp - lagmatrix(temp,1)==1, 1);
    SendDates = temp(tempPos-1);
    ReturnDates = LVTSdates(tempPos);
    FauxLoans(:,1) = SendDates;
    FauxLoans(:,11) = ReturnDates;
    TrainingData = [FurfineData; FauxLoans];
end

    
    
            
    
    
