#! R

##################################################################################################
# Script that applies the Furfine algorithm and/or variants to the resampled days.
#
# Inputs: 1) An .RData file (generated by BootstrapSampling.R) containing the payments data with synthetic 
#               date labels such that days for which the algorithm should be run on are adjacent business days
#               and otherwise the date labels should be spaced greater than one business day
#         2) The file name (RData) in which to save the output (i.e. false results from Furfine algorithm)
#         3) The variant of the Furfine algorithm to run: `Hendry' (is Baseline algorithm with closest to target duplicate treatment)
#                                                         `Rempel' (is kNN refined variant)
#         4) Indicator (1/0) whether to open a matlabpool and run the algorithm in parallel in matlab
#         5) If Parallel = 1, indicates number of matlab workers to open (if 0, number is just a placeholder)
#         6) specify minimum send time of payments data 
#         7) minimum loan size in dollars (e.g. 10^6 = 1 million)
#         8) interest rate threshold, lower limit deviation from target (e.g. 0.5 is 50 basis points below target rate)
#         9) interest rate threshold, upper limit deviation from target (e.g. 0.5 is 50 basis points below target rate)  
#         10) refinement can be `None', 'IntRound', 'IntRoundExact', 'EndOfDay' 
#               -IntRoundExact ensures all loans identified have rounded interest rates with an additional check that the repayments conform to the rounded rate
#               -EndOfDay restricts consideration of all payments on considered lend day of 4pm or afterwards (24 hour clock used in numerics, i.e. 4pm = 160000)
#       If providing more than 10 arguments, then presumed using the kNN refinement, see 
#
# There is also an implicit dependency on the working directory containing FurfineControlScript.m which is the code that implements on specified variants of Furfine given proper inputs
#
# Output is an RData file with object Results
#     Results[[i]] = list(bootNumber=b, num_loans=numloans, matches=Fresults, comp.time = comp.time, num_nonunique = num_nonunique, num_obs=nrow(my.sample))
#       #  Results[[i]]$matches gives the falsely identified Furfine matches for the i-th lend date
#       # Results will have the same length as BootSample
# Rempel, May, 2014.
###################################################################################################


# RempelBootTest.R --args 'BootstrapSample.RData', 'OutputName.RData', 'Rempel', Parallel, nMatSlaves,  min_sentime, min_loansize, bpsl, bpsh, refinement, SyntheticTraining, Control, useDANN, kloss, choiceK, Xpaythreshold, methodDup2, training prefix str, training suffix string, lvts data file (.mat)

###########################
## Initialization


print('starting Rempel Bootstrap Test')
#pathToScripts <- '/home/mfa/remm/Replication'  
pathToScripts <- getwd()
# Dependencies:
source(paste(pathToScripts, '/InitRmatlabFunctions.R',sep=''))
source(paste(pathToScripts, '/RempelBootstrapTest_SlaveFunction.R', sep=''))
FurfineScript <- 'FurfineControlScript' # matlab matching algorithm m-file

# Global parameters
portnum <- 9999 # local host port number to be used for R - matlab connection
Seed <- 205030

# Obtaining Script Parameters

args <- commandArgs(trailingOnly = TRUE)
if(length(args)>21) portnum <- args[22] # if port number is supplied as argument, override default
print(paste('Port number to be used for local R-matlab connection is ', portnum , sep=' '))
inputfilename <- args[1]
outputfilename <- args[2]
sprintf('for output %s', outputfilename)
methodTreatDuplicates <- args[3]
Parallel <- as.numeric(args[4])
nMatSlaves <- as.numeric(args[5])
min_sentime <- as.numeric(args[6])
min_loansize <- as.numeric(args[7])
bpsl <- as.numeric(args[8])
bpsh <- as.numeric(args[9])
refinement <- args[10]
endofdaythreshold <- 160000 # NOTE End of day threshold hardcoded to 4pm for EndOfDay Refinement
if(refinement=='EndOfDay') print(paste('End of day refinement to be used, with threshold of ', endofdaythreshold))


Simulation <- 1
Nchunks = 1;
print(refinement)
print(min_sentime)
print(min_loansize)
if(length(args) > 10){
	# then it is the additional options for kNN Criterion
	SyntheticTraining <- as.numeric(args[11])
	ControlSyntheticForParticipants <- as.numeric(args[12])
        useDANN <- as.numeric(args[13])
        klossWeighted <- as.numeric(args[14])
        choiceK <- as.numeric(args[15])
        Xpaythreshold <- as.numeric(args[16])
	methodDup2 <- args[17]
	training_suffix_str <- args[19]
	training_prefix_str <- args[18]	
	ModelTraining <- as.numeric(args[20])
	lvtsdatafile <- args[21]
	useReturnLegCharacteristics <- 1
}
jobname <- sub('.RData', '', outputfilename)

# Check if system has MPI capabilities and run job in serial / parallel accordingly
num_cores <- NULL
try({num_cores <- mpi.comm.size()},silent=T )

if(is.null(num_cores) || num_cores <= 1){
        # Will run job in serial
        print('Parallel computing facilities not found. Running job as serial')
	rank <- 0
	success <- NULL
	OUT <- try({openLocalRmatlabConnection(port=portnum)}, silent=T)
	try({success <- OUT$success}, silent=T)
	try({matlab <- OUT$matlab}, silent=T)
	
	if(success==1 || inherits(OUT,'try-error') || is.null(success)){
		# matlab connection failed
		close(matlab)
		stop('Failed to properly open matlab connection. Aborting. \n Check that there no conflicts with port used for R-matlab connection and that R.matlab is properly set-up')
		
	} else {
		print("Succesfully opened R-matlab connection.")
	}


	setVariable(matlab,rank=0)
        (evaluate(matlab, "fprintf('[%d] Getting parameters', rank);"))
        (setVariable(matlab, Nchunks=Nchunks,
                Parallel=Parallel, nMatSlaves=nMatSlaves, min_sentime = min_sentime,
                min_loansize=min_loansize, methodTreatDuplicates=methodTreatDuplicates,
                refinement=refinement, bpsh=bpsh, bpsl=bpsl, Simulation=Simulation,Seed=Seed,
                output_string = outputfilename
        ))
        if(length(args) > 10){
                    (setVariable(matlab, Xpaythreshold=Xpaythreshold,
                    choiceK=choiceK,
                    ControlSyntheticForParticipants=ControlSyntheticForParticipants,
                    useReturnLegCharacteristics=useReturnLegCharacteristics,
                    useDANN=useDANN,
                    klossWeighted=klossWeighted,
                    ModelTrainingData=ModelTraining,
                    methodDup2=methodDup2,
                        lvtsdatafile = lvtsdatafile
                ))
        }
       if(refinement=='EndOfDay'){
                (setVariable(matlab,endofdaythreshold=endofdaythreshold))
        }
        (evaluate(matlab, "addpath('KNN'); \n"))
	 (evaluate(matlab, "rng(Seed); \n"))

	#evaluate(matlab, "rng(Seed); \n") # setting seed in matlab session in case of sampling
        print('Initialization Complete')
        load(inputfilename)

        (print("Main Function About To Start"))

        # Master Function

        numboots <- length(BootSample)
        tasks <- vector('list', length=numboots)
        tasks <- lapply(1:numboots, function(x) list(bootNumber=x, bootsample=BootSample[[x]]$bootsample))
        master.last.slave.time <- proc.time()[3] + 250
	Results <- lapply(tasks, RempelTest)
	sprintf("Job has run it's course. Saving Results in %s", outputfilename)
        save(Results, file=paste('Results/', outputfilename, sep=''))
        print('Saved. Exiting.')
	close(matlab)
	cat('closed matlab')
} else {
        # Will run job in parallel
	#  Open log files on slaves and set parallel random number streams
	
        #require(rlecuyer)
	print("Parallel facilities found, implementing parallel testing variant.")

        mpi.setup.rngstream(iseed=Seed)
	mpi.bcast.cmd(pSeed <- sample(1:50000,1))
	# distribute parameters to workers
	mpi.bcast.Robj2slave(jobname)

	############
	# Initialize R-Matlab on each core
	success <- mpi.remote.exec(try({
	    out_fname <- sprintf("%s_output_slave.%03d.txt", jobname, mpi.comm.rank())
	    zz <- file( out_fname, open="wt")
	    sink( zz, append=TRUE )
	    sink( zz, type="message", append=TRUE )

	    require(R.matlab);
	    matlab.port <- as.integer(Sys.getenv("MATLABSERVER_PORT"))
	    cat(sprintf("rank %03d will use matlab port %5d\n", mpi.comm.rank(), matlab.port))
	    matlab <- Matlab(port=matlab.port)
	    open(matlab)
	    success <- isOpen(matlab)
	    if(success) {
		setOption(matlab, "readResult/interval", 15)  # in seconds
		setOption(matlab, "readResult/maxTries", 120)  # 15 sec times 60 tries = 15 minutes
		matlab.diary <- sprintf("%s_output_matlab.%03d.txt", jobname, mpi.comm.rank())
		if (file.exists(matlab.diary)) file.remove(matlab.diary)
		evaluate(matlab, paste("diary ", matlab.diary ))
		evaluate(matlab, paste("rank = ", mpi.comm.rank(), ";"))
	    }
	    success
	}), ret=TRUE)

	if (any(sapply(success, inherits, "try-error")) || !all(as.logical(success))) {
	    mpi.bcast.cmd( close(matlab) )
	    mpi.close.Rslaves()
	    mpi.quit()
	}




	############


		mpi.bcast.Robj2slave(Simulation)
		mpi.bcast.Robj2slave(endofdaythreshold)

	mpi.bcast.Robj2slave(nMatSlaves)
	mpi.bcast.Robj2slave(FurfineScript)
	if(length(args) > 10){
		useReturnLegCharacteristics <- 1
		mpi.bcast.Robj2slave(SyntheticTraining)
		mpi.bcast.Robj2slave(useReturnLegCharacteristics)
		mpi.bcast.Robj2slave(ControlSyntheticForParticipants)
		mpi.bcast.Robj2slave(training_suffix_str)
		mpi.bcast.Robj2slave(training_prefix_str)
		mpi.bcast.Robj2slave(useDANN)
		mpi.bcast.Robj2slave(klossWeighted)
		mpi.bcast.Robj2slave(outputfilename)
		mpi.bcast.Robj2slave(Xpaythreshold)
		mpi.bcast.Robj2slave(ModelTraining)
		mpi.bcast.Robj2slave(choiceK)
		mpi.bcast.Robj2slave(methodDup2)
		mpi.bcast.Robj2slave(lvtsdatafile)
	}

	mpi.bcast.Robj2slave(nMatSlaves)
	mpi.bcast.Robj2slave(Parallel)
	mpi.bcast.Robj2slave(Nchunks)
	mpi.bcast.Robj2slave(methodTreatDuplicates)
	mpi.bcast.Robj2slave(min_sentime)
	mpi.bcast.Robj2slave(min_loansize)
	mpi.bcast.Robj2slave(bpsl)
	mpi.bcast.Robj2slave(bpsh)
	mpi.bcast.Robj2slave(refinement)
	mpi.bcast.Robj2slave(outputfilename)
	mpi.bcast.Robj2slave(pathToScripts)
	# Setting parameters in the matlab worker workspaces:

	mpi.bcast.cmd(print('Passing parameters to mat workers\n'))
        mpi.bcast.cmd(source(paste(pathToScripts, '/RempelBootstrapTest_SlaveFunction.R', sep='')))

	mpi.bcast.cmd(evaluate(matlab, "fprintf('[%d] Getting parameters', rank);"))
	mpi.bcast.cmd(setVariable(matlab, Nchunks=Nchunks,
		Parallel=Parallel, nMatSlaves=nMatSlaves, min_sentime = min_sentime,
		min_loansize=min_loansize, methodTreatDuplicates=methodTreatDuplicates,
		refinement=refinement, bpsh=bpsh, bpsl=bpsl, Simulation=Simulation,
		output_string = outputfilename
	))

	if(length(args) > 10){
		    mpi.bcast.cmd(setVariable(matlab, Xpaythreshold=Xpaythreshold,
		    choiceK=choiceK,
		    ControlSyntheticForParticipants=ControlSyntheticForParticipants,
		    useReturnLegCharacteristics=useReturnLegCharacteristics,
		    useDANN=useDANN,
		    klossWeighted=klossWeighted,
		    ModelTrainingData=ModelTraining,
		    methodDup2=methodDup2,
			lvtsdatafile = lvtsdatafile
		))
	}

	if(refinement=='EndOfDay'){
		mpi.bcast.cmd(setVariable(matlab,endofdaythreshold=endofdaythreshold))
	}
	mpi.bcast.cmd(evaluate(matlab, "addpath('KNN'); \n"))
	mpi.bcast.cmd(setVariable(matlab,Seed=pSeed))
	mpi.bcast.cmd(evaluate(matlab, "rng(Seed);"))
	mpi.bcast.cmd(evaluate(matlab, "test=Nchunks;"))
	mpi.bcast.cmd(testR <- getVariable(matlab,"test"))
	mpi.bcast.cmd(print(paste('test Mat output', testR)))

	print('Initialization Complete')
	load(inputfilename)

	mpi.bcast.cmd(print("Main Function About To Start"))


	# tags sent by the master to the slaves
	master.tag.work <- 1
	master.tag.done <- 2
	# tags sent by the slaves to the master
	slave.tag.ready <- 101
	slave.tag.finished <- 102
	slave.tag.quit <- 103

	mpi.bcast.cmd({ master.tag.work <- 1; master.tag.done <- 2;
			slave.tag.ready <- 101; slave.tag.finished <- 102; slave.tag.quit <- 103})

	mpi.bcast.Robj2slave(slavefunction)
	mpi.bcast.cmd(slavefunction())

	# Master Function

	numboots <- length(BootSample)
	tasks <- vector('list', length=numboots)
	tasks <- lapply(1:numboots, function(x) list(bootNumber=x, bootsample=BootSample[[x]]$bootsample))
	#print(tasks[[numboots]])
	Results <- vector('list', length=numboots)
	junk <- 0
	closed_slaves <- 0
	n_slaves <- mpi.comm.size() - 1
	print(n_slaves)
	master.last.slave.time <- proc.time()[3] + 250
	while(closed_slaves < n_slaves){
		# receive any message from any slave
		proc.time()[3] -> master.initial.time
		slave.msg <- mpi.recv.Robj(mpi.any.source(), mpi.any.tag())
		slave.msg.info <- mpi.get.sourcetag()
		slave_id <- slave.msg.info[1]
		slave_tag <- slave.msg.info[2]
		print(paste('slave id', slave_id, ' slave_tag ', slave_tag))
		if (slave_tag==slave.tag.ready) {
			if(length(tasks) > 0){
				this.task <- tasks[[1]]
				tasks[1] <- NULL
				print(paste('Number of tasks left -', length(tasks)))
				mpi.send.Robj(this.task, slave_id, tag=master.tag.work);
			} else mpi.send.Robj(obj=NULL, dest=slave_id, tag=master.tag.done)
		} else if (slave_tag==slave.tag.finished) {
			# then we're receiving results
			print(paste('Received Results', 'Subbing them into Results', sep='\n'))
			Results[[slave.msg$bootNumber]] <- slave.msg


		} else if (slave_tag==slave.tag.quit) {
			# A slave has shut down
			print('Slave has closed down')
			closed_slaves <- closed_slaves + 1
			if (n_slaves - closed_slaves==1){
				proc.time()[3] -> master.last.slave.time
			}
		}
		flush.console()
		if(n_slaves - closed_slaves==1){
			if(master.last.slave.time + 120 < proc.time()[3]){
				print('Master has hung-up on the last task')
				break
			}
		}
		if(master.initial.time*100 < proc.time()[3]){
		       print(paste("Master is presumed to have hung-up"))
		       break
		}
	} # All tasks complete across slaves

	sprintf("Job has run it's course. Saving Results in %s", outputfilename)
	save(Results, file=paste('Results/', outputfilename, sep=''))

	print('Saved')

	cat("closing matlabs\n")
	mpi.bcast.cmd(close(matlab))

	mpi.close.Rslaves()
	mpi.quit()
		
}
