#! R

# This script generates simulated (dummy) payment data to be used to illustrate the scripts needed to replicate
	# Rempel (2014) 

# The payments data will contain
	# 1) uniformly distributed non-loan payments for 4 participants 
	# 2) overnight loans between the participants 
	# 3) simulated FX payments

# data will be in form
	# date, time, value, sender, receiver, FXrate, ONrate, FXpaymentIndicator, ONloanIndicator, transnum

# data will be outputed into Inputs sub-folder.
 

args <- commandArgs(trailingOnly = TRUE) # take the arguments trailing the Rscript in batch submission as inputs
#residargs <- ls()
#residargs <- residargs[-grep('args', residargs)]
#rm(list=residargs)
outputfilename <- args[1] 
basedate_simulation <- as.numeric(args[2]) # assuming in form - yyyymmdd - where dd %in% c(01, 02, 03) (to ensure all generated dates are valid with as.Date() )
Seed <- as.numeric(args[3]) 
print(args[1])
print(args[2])
print(Seed)
set.seed(Seed)

source('SimulatePaymentsFunctions.R')

# Simulation Parameters, path of ON/FX Rates / Dates

BoCid <- 2 # participant code for Bank of Canada (non-Bank LVTS participant)
num_participants <- 4
num_nonloanpayments_perparticipant <- 10^4
payment_mean.per.participant <- 4*10^7/num_participants # roughly 40 million total transactions
payment_sd.per.participant <- 10^7
num_FX <- 50
num_ON <- 4
# generating 4 weeks of dates

date_range <- c(20200101:20200105, 20200108:20200112)
FXrates <- rep(round(seq(1, 1.8, length=length(date_range)/2),digits=4), each=2)
ONrate <- 2.5
avgONsize <- 3*10^7
sdONsize <- 10^7
ratesData <- cbind(date_range, FXrates, rep(ONrate, length(date_range))) # NOTE FX rate for each day must be lagged by T-1 or T-2 depending on settlement timing of FX for the FX market in question
colnames(ratesData) <- c('date', 'FXrate', 'ONrate')
date_range2 <- c(20200115:20200119, 20200122:20200126)
ONrate2 <- 2
ONrates2 <- c(ONrate, ONrate, rep(ONrate2, length(date_range2) - 2))
ratesData2 <- cbind(date_range2, FXrates, ONrates2) 

ratesData <- rbind(ratesData, ratesData2)
ratesData <- as.data.frame(ratesData)
save(ratesData, file='Inputs/SimulatedFXandONrateData.RData')
full_date_range <- c(date_range,date_range2)

# loop through date_range generating each type of date per day

payments_data.l <- vector('list', length(full_date_range))
loan_data.l <- vector('list', length(full_date_range))
base_transnum <- 1
for(day in full_date_range){
	
	ONrate.day <- ratesData$ONrate[which(day==full_date_range)]
        FXrate.day <- ratesData$FXrate[which(day==full_date_range)]
        FXrate.nextday <- ratesData$FXrate[which(day==full_date_range)+1]
	nonloan.day <- generate_dailynonloan_paymentflows(num_participants, num_nonloanpayments_perparticipant, payment_mean.per.participant, payment_sd.per.participant,ONrate.day, FXrate.day,  day)
	FX.day <- generate_FXpayments(num_participants, num_FX, ONrate.day, FXrate.day,  day)

	if(!day==max(full_date_range)){
		ON.day <- generate_overnightloans(num_participants, num_ON, ONrate.day, FXrate.day, FXrate.nextday, avgONsize, sdONsize, day, full_date_range[which(day==full_date_range)+1])
		loans <- ON.day$trueloans
		if(length(which(loans[,12] - loans[,3] <0)) > 0) stop('Repayment smaller than principal') 
		loan_data.l[[which(day==full_date_range)]] <- ON.day$trueloans
		ON.day <- ON.day$payoutput
		
	} else {
		ON.day <- NULL
	}
	pay.day <- rbind(nonloan.day, FX.day, ON.day)
	colnames(pay.day) <- c('date', 'time', 'value', 'sender', 'receiver', 'FXrate', 'ONrate', 'FXind', 'ONind')
	pay.day <- pay.day[order(pay.day[,1], pay.day[,2]),]
	transnum.day <- seq(base_transnum, base_transnum + nrow(pay.day) -1,1)
	ON.transnum.day <- transnum.day
	pay.day <- cbind(pay.day, transnum.day)
	base_transnum <- max(transnum.day) + 1
	stopifnot(length(which(is.na(pay.day)))==0)
	payments_data.l[[which(day==full_date_range)]] <- pay.day
}

payments.data <- do.call(rbind, payments_data.l)
payments.data <- as.data.frame(payments.data)

payments.data$sender[which(payments.data$sender==BoCid)] <- num_participants + 1
payments.data$receiver[which(payments.data$receiver==BoCid)] <- num_participants + 1




save(payments.data, file=outputfilename)
write.table(as.matrix(payments.data), file=sub('.RData', '.csv', outputfilename), row.names=F, col.names=F, sep=',')

rlvts <- payments.data
rlvts$date <- as.Date(as.character(rlvts$date), format='%Y%m%d')
save(rlvts, file=sub('.RData', '_rlvts.RData', outputfilename))


loan.data <- do.call(rbind, loan_data.l)
loan.data <- as.data.frame(loan.data)
lrates <- (loan.data$V12 - loan.data$V3)/loan.data$V3*100*365/(as.numeric(as.Date(as.character(loan.data$V10), format='%Y%m%d') - as.Date(as.character(loan.data$V1), format='%Y%m%d')))
loan.data <- cbind(loan.data, lrates)

save(loan.data, file=paste(sub('.RData', '', outputfilename),  '_trueloandata.RData', sep=''))

write.table(as.matrix(loan.data), file=paste(sub('.RData', '', outputfilename),  '_trueloandata.csv', sep=''), row.names=F, col.names=F, sep=',')

