#!R 
# Script contains functions to compute the positive false discovery rates
#  Main function is SummarizeTypeIerror()
#	pfdr_estimate() computes the pfdr for a given sample of a single lend date re-paired with (multiple) repayment dates
#	extractNumLoans - takes the date supplied and determines the number of furfine-variant identified transactions on correctly ordered payments data
#	QuantilesTableG() - computes summary stats of the pfdr's (quantiles exceeding threshold percentage levels - e.g. percent exceeding 20\%)
#	DuplicatePercentageTable() - compares the percentage of non-uniquely identified furfine-transactions from a list of inputted data
#	getDates() - relabels the dates of the identified loans from the resampled data to their actual transaction dates






pfdr_estimate <- function(results.l, NumBoot.l, Days.l, AllDays.l){
	# for a given sender date, compute the pfdr
		# this function takes the false matches given in results.l, the number of loans identified on the correct day orderings (NumBoot.l) for the sampled lend day (lend day in results.l), 
			# AllDays.l lists the sampled return payment days matched with the lend day. 
                datecol <- 11 # return leg transaction date column from matched data
		sdatecol <- 1 # principal leg transaction date column from matched data
		if(length(AllDays.l)>0) B_R <- length(AllDays.l) else B_R <- 20 # Default case of number of return days is 20

		dayspl <- vector('list', B_R)
		for(day.indx in 1:B_R){
			dayspl[[day.indx]] <- results.l[which(results.l[,datecol]==AllDays.l[day.indx]),]
		}
		numfalsepositives <- do.call(c,lapply(1:length(dayspl), function(z) nrow(dayspl[[z]]))) # number of matches of alg. on resampled payment data
		numfalsepositives[which(is.na(numfalsepositives))] <- 0 # replacing missing with 0 (which would have resulted from no spurious matches for a sample)
                Ediscovery <- mean(numfalsepositives, na.rm=T) 
		Pr_discovery <- length(which(numfalsepositives>0))/B_R # number of sampled return days which generated at least 1 match / number of sampled return days
                num_furfloans <- NumBoot.l[which(results.l[1,sdatecol]== Days.l)] # getting the number of loans from the sender date
		if(length(num_furfloans)==0 || num_furfloans==0) num_furfloans <- 1 # if no loans identified on this day
		if(Pr_discovery==0){
			pfdr <- 0
		} else {
	                pfdr <- Ediscovery/(Pr_discovery*num_furfloans)
		}
                return(pfdr) # is B_R x 1
}



extractNumLoansInDay <- function(X.l, furfdat){
	# this function takes a list of bootstrap payment or repayment dates and returns the number of loans identified by furfine on the corresponding date(s)
        # X.l is the list of bootstrap days
	# assumes the dates withing X.l[[1]] are in form 'yyyy-mm-dd'
	if(nchar(as.character(X.l[[1]]))!= 10) stop('Format of dates not as anticipated. Ensure of form "yyyy-mm-dd"') 
        numFurfLoansPerDayVector <- do.call(c,lapply(1:length(X.l), function(x) nrow(furfdat[which(furfdat[,1]==as.numeric(gsub('-', '', as.character(X.l[[x]])))),])))
        return(numFurfLoansPerDayVector)
}






QuantilesTableG <- function(y){ 
	# table of bootstrap test statistics (FDRs) greater than 20%/10%/5%/1% of time
	numbootobs <- length(y)
	mean_t <- mean(y, na.rm=T)
	p_value_20 = length(which(y[1:numbootobs]>0.2))/numbootobs
	p_value_10 = length(which(y[1:numbootobs]>0.1))/numbootobs
	p_value_05 = length(which(y[1:numbootobs]>0.05))/numbootobs
	p_value_01 = length(which(y[1:numbootobs]>0.01))/numbootobs
	Bootteststats <- c(mean_t, p_value_20, p_value_10, p_value_05, p_value_01)
	names(Bootteststats)[1:5] <- c('mean(\tau)', 'Prob(\tau > 20%)','Prob(\tau > 10%)','Prob(\tau > 5%)', 'Prob(\tau > 1%)')
	return(Bootteststats)
}


DuplicatePercentagesTable <- function(results){
	# Generates Table 2 
		# - takes a list of dataframes, and computes percentage which are generated by duplicate matches
	percentdups <- sapply(results, function(results.l) length(which(results.l[,22]>0))/nrow(results.l))
	lq <- quantile(percentdups, c(0.01, 0.25, 0.5))
	uq <- quantile(percentdups, c(0.75, 0.99))
	meanpd <- mean(percentdups)
        SumP <- c(lq, meanpd,uq)
        names(SumP)[c(4)] <- c('Mean')
        return(SumP)
}

getDates <- function(x, results, lvts, transIDcol, datecol){
		# takes the identified overnight loan transactions from rematched data, and relabels the transaction of the legs to their real transaction dates 
                b <- results[[x]]
                sublvts <- lvts[,c(1,10)]
		if(class(b)=='matrix'){
			b <- as.data.frame(b)
			names(b) <-  c('date', 'time', 'value', 'sender', 'receiver', 'tranche', 'ONrate', 'jumbo', 'misc', 'transnum', 'date_r', 'time_r', 'value_r', 'sender_r', 'receiver_r', 'tranche_r', 'ONrate_r', 'jumbo_r', 'misc_r', 'transnum_r', 'furf_rate', 'mltpl')

		} 
                b2 <- merge(b, sublvts, by.x=names(b)[transIDcol], by.y=names(sublvts)[2]) # using the transaction identifier column 
                b2 <- b2[,c(1,ncol(b) +1)]
                b <- b[order(b[,transIDcol]),]
                b[,datecol] <- b2[,2] # replacing the artificial date labels used for running Furf algorithm to their actual dates
                return(b)
}




## Main function:


SummarizeTypeIerror <- function(Results.filename, BootData.filename, FurfData.filename, lvts, rlvts, teststat){
	# takes the Permutation Results object containing the Furfine-variant algorithm output from repaired data in list form
	#	the resampled payments data and the original Furfine-variant output (on the correctly ordered payments data)
	#	and computes the pfdr for each lend date used as a base
	#
	#	First obtains the actual date labels from the resampling
	#	relables the Results data dates to correspond to the payments true transaction date
	#	extracts the number of loans identified on the original application of the algorithm for each of the sampled loan / return days
	#	computes the pfdr and stores it in tau 
	#	

	load(Results.filename) # yields object Results
	load(BootData.filename) # yields object BootSample
	load(FurfData.filename) # yields furfdat object


	# furfdat (furfine data)
	colnames(furfdat) <- c('date', 'time', 'value', 'sender', 'receiver', 'tranche', 'ONrate', 'jumbo', 'misc', 'transnum', 'date_r', 'time_r', 'value_r', 'sender_r', 'receiver_r', 'tranche_r', 'ONrate_r', 'jumbo_r', 'misc_r', 'transnum_r', 'furf_rate', 'mltpl')
	furfdat <- as.data.frame(furfdat)
	# ensuring no loans outside the standard window
	furfdat <- furfdat[which(furfdat$time >= 60000 & furfdat$time_r >= 60000),]
	DaysReturned.boot <- lapply(BootSample, function(b) unique(b$SRdates[,2]))
	DaysLent.boot <-lapply(BootSample, function(b) unique(b$SRdates[,1]))

	AllDaysReturned.boot <- lapply(BootSample, function(b) b$SRdates[,2])
	AllDaysLent.boot <-lapply(BootSample, function(b) b$SRdates[,1])

	NumLoansOnReturnedBootDay.l <- lapply(DaysReturned.boot, function(x) extractNumLoansInDay(x, furfdat)) # number of loans on repayment day as base
			# above not used in computation of pfdr
	NumLoansOnLendBootDay.l <- lapply(DaysLent.boot, function(x) extractNumLoansInDay(x, furfdat)) # number of loans on each lend day


	results <- lapply(Results, function(x) x$matches) # get the algorithm matched transactions from the resampled data
	results <- lapply(results, function(x) if(is.null(x)) matrix(0, ncol=22, nrow=1) else x) # if no spurious matches, make a zero dimension matrix

	# get the true dates for the bootstrap sample

	resultsR <- lapply(1:length(results), function(x) try(getDates(x, results, lvts, transIDcol=20, datecol= 11)))

	resultsSR <-  lapply(1:length(results), function(x) try(getDates(x, resultsR,lvts,  10, datecol= 1)))
	pfdrstats.l <- lapply(1:length(resultsSR), function(x) try(pfdr_estimate(resultsSR[[x]], NumLoansOnLendBootDay.l[[x]], DaysLent.boot[[x]], AllDaysReturned.boot[[x]])))


	pfdrstats <- do.call(c, pfdrstats.l)
   	       if(teststat=='pfdr'){
                        tautilde <- pfdrstats
                } else {
   			print('statistic pfdr not specified, alternatives not supported')
		}
		typeIstats <- tautilde
	
	typeItable <- QuantilesTableG(typeIstats)
	typeIduplicateTable <- DuplicatePercentagesTable(resultsR)
	

	typeIepdf <- NULL
try(	typeIepdf <-density(typeIstats))

	typeIoutput <- list(tau = typeIstats,qtable = typeItable, duptable = typeIduplicateTable, tau_density = typeIepdf, raw = resultsSR)
	return(typeIoutput) 	
}
	
replaceSyntheticDates <- function(indx, results.l, daysReturned.l, lvts){
	# this function replaces the artificial dates in the bootstrap data,
		# with the actual dates which were sampled
		# matching on the basis of transaction ID numbers

	daysReturned <-  daysReturned.l[[indx]]
	sublvts <- lvts[which(lvts[,1] %in% daysReturned),]
	results.b <- results.l[[indx]]
	daysReturnedResultsIndx.l <- lapply(daysReturned, function(x) which(results.b[,20] %in% lvts[which(lvts[,1] == x),10]))
	for(i in 1:length(daysReturned)){
		results.b[daysReturnedResultsIndx.l[[i]],11] <- daysReturned[i]
	}
	return(results.b)
} 


